from functools import wraps
from time import time

indent = 2
space = ' '


def timing(f):
    @wraps(f)
    def wrap(*args, **kw):
        global indent, space
        print(f'{space: <{indent}}func:{f.__name__} args:[{args}, {kw}] started')
        indent += 2
        ts = time()
        result = f(*args, **kw)
        te = time()
        indent -= 2
        print(f'{space: <{indent}}func:{f.__name__} finished: {te-ts:2.4f} sec')
        return result
    return wrap



#!/usr/bin/env python3
import os
import sys
import argparse
from argparse import RawTextHelpFormatter

from .utils.model import model
from .utils.dwhrepo import repo
from .utils.log import logger, init_logger, headline
from .utils.release import release

JOBS = {}

def register_action(function):
    JOBS[function.__name__] = function.__doc__
    def wrapper(args):
        headline(function.__doc__)
        function(args)
        headline("End {}".format(function.__name__), True)
    return wrapper

@register_action
def prepare(args) -> int:
    """Prepares release candidate file."""
    model.get_all_ddls()
    model.prepare_release_candidate(force=args.force)

@register_action
def deploy(args):
    """Deploys changes from release candidate file."""
    model.deploy_release(dry_run=args.dry_run)

@register_action
def migrate(args):
    """prepare + deploy """
    prepare(args)
    deploy(args)

@register_action
def validate(args):
    """Validates all .sql files in model directory"""
    model.get_all_ddls()

@register_action
def history(args):
    """Prints release history."""
    release.print_release_history()

@register_action
def clone(args):
    """Clones (or replaces) database based on prod."""
    model.clone_production(force=args.force)

@register_action
def sync(args):
    """Syncs unapplied changes from model and releases dirs."""
    release.sync(dry_run=args.dry_run)

@register_action
def test_sync(args):
    """Test release on a separate clone (run it before creating pull request)."""
    release.test_sync()

@register_action
def compare(args):
    """Compares Snowflake and current branch DDLs."""
    if args.file:
        model.compare_single_file(filename=args.file.name)
    else: 
        model.compare_sf_git()

@register_action
def diff(args):
    """Prints diff from production."""
    repo.diff_from_prod()

@register_action
def abandoned(args):
    """Compares active branches and development clones."""
    release.compare_branches_and_clones()

def main():
    jobs = list(JOBS.keys())

    description = """  Git <-> Snowflake sync and automatic deployment. See
  README.md file for full documentation.

Actions:\n\n"""
    for job in jobs:
        description += f"  {job:10}\t\t{JOBS[job]}\n"

    parser = argparse.ArgumentParser(add_help=True,
            prog="cicd",
            formatter_class=RawTextHelpFormatter,
            description=description)
    parser.add_argument("-v", "--verbose", help="Verbose mode. Shows SQL "
                        "statements.", action="store_true")
    parser.add_argument("-t", "--dry-run", help="Show SQL to be executed, "
                        "but doesn't run it.", action="store_true")
    parser.add_argument("-f", "--force", help="Force command without yes/no "
                        "question asked in terminal.", action="store_true")
    parser.add_argument("action",  nargs='+', help="Action to run", choices=jobs)
    parser.add_argument("--file", action="store", help="Filename for compare action",
                        type=argparse.FileType('r'), nargs='?')

    if len(sys.argv)==1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()
    init_logger(args)

    try:
        for action in args.action:
            globals()[action](args)
    except (RuntimeError, AssertionError) as e:
        logger.error(e)
        quit(-1)

if __name__ == "__main__":
    main()



if __name__ == '__main__':
    try:
        from cicd.utils.log import logger, init_logger
        init_logger()
        from cicd.cicd import main
        main()
    except (RuntimeError, AssertionError) as e:
        logger.error(e)
        quit(-1)
    


import logging

import coloredlogs
from termcolor import colored

logger = logging.getLogger("snowflake.cicd")

def init_logger(args=None):
    coloredlogs.install(
            level='DEBUG' if args and args.verbose else 'INFO',
            logger=logger,
            fmt='%(asctime)s %(programname)s %(message)s',
            programname='>',
            datefmt='%H:%m:%S',
            field_styles= {
                    'asctime': {'color': 'black', 'bright': True},
                    'hostname': {'color': 'magenta'},
                    'levelname': {'bold': True, 'color': 'black'},
                    'name': {'color': 'blue'},
                    'programname': {'color': 'black', 'bright': True},
                    'username': {'color': 'yellow'}},
            level_styles= {
                    'critical': {'bold': True, 'color': 'red'},
                    'debug': {'color': 'white', 'faint': True},
                    'error': {'color': 'red'},
                    'info': {'color': 'white', 'faint': False},
                    'notice': {'color': 'magenta'},
                    'warning': {'color': 'yellow'}})

    logging.getLogger("snowflake.connector.network").disabled = False
    logging.getLogger("snowflake.cicd").level = logging.DEBUG if args and args.verbose else logging.INFO

def is_debug():
    return logger.level <= logging.DEBUG

def headline(msg, end=False):
    if end:
        logger.info(colored("{:>50}  ".format(msg), 'white',
                attrs=['reverse', 'dark']))
    else:
        logger.info(colored("  {:50}".format(msg), 'white',
                attrs=['reverse']))



import shutil
import os
import re
from pathlib import Path

from .log import logger, is_debug
from .dwhrepo import repo
from .snowflake import sf
from .release import release
from .sql import print_sql, sql_meta, get_diff_sql, statement_cleanup
from .utils import get_file_contents
from .config import config


class Model():
    """Handles model .sql definitions."""

    MODEL_DIR    = config.read_config('model_dir', default='model')
    EXTENSIONS   = re.compile(r'.*(\.sql)|(\.vw)|(\.tbl)$', re.I)
    DIFF_DIR     = ".diff"

    def __init__(self):
        self.sf_safe_branch  = repo.get_sf_safe_branch()

    def prepare_release_candidate(self, force):
        """Prepares a release candidate file (if missing)."""
        repo.assert_repo()
        commit_hash = release.get_base_commit(self.sf_safe_branch)
        release.check_release_dir_clean(base_commit=commit_hash)

        files = repo.get_changed_files(index=commit_hash,
                prefix=Model.MODEL_DIR)

        if len(files) == 0:
            logger.info(f"No changes in {Model.MODEL_DIR} dir to prepare release candidate "
                    "file. Preparing empty file.")

        if force:
            logger.info("Not checking if release candidate file was changed due to --force.")
        else:
            release.check_release_candidate()

        branch = repo.get_branch()

        release_candidate_sql = f"--.Release candidate file, branch: {branch}\n\n"
        for change in files.values():
            release_candidate_sql += Model._change_into_release_entry(change, commit_hash)

        release.save_release_candidate_file(release_candidate_sql, branch)
        logger.debug("Release candidate file contents:\n" + release_candidate_sql)

    def deploy_release(self, dry_run=True):
        release_sql = release.prepare_release_file()
        deploy_sql = release.release_file_to_sql(release_sql)

        if is_debug():
            print_sql(deploy_sql)
        if dry_run:
            if not is_debug():
                print_sql(deploy_sql)
            logger.info("Skipping SQL execution due to --dry-run.")
            return

        sf.perform_release(deploy_sql, self.sf_safe_branch)
        release.save_release(release_sql)

        pass
        
    def clone_production(self, force):
        """Clones production database into new db named after branch name."""
        sf.clone_production(self.sf_safe_branch, force)

    def compare_sf_git(self, branch=None):
        """Compares Snowflake and current branch DDLs."""

        if branch is None:
            branch = self.sf_safe_branch

        git_ddls = self.get_all_ddls()
        sf_ddls = sf.get_all_objects(branch)

        logger.info("| {:_^39.39} | {:_^11.11} | {:_^30.30} | {:_^16.16} |".format(
           'object name', 'type', 'GIT file name', 'last change on SF'))   

        for name in sorted(git_ddls.keys() | sf_ddls.keys()):
            o_name = git_ddls.get(name, sf_ddls.get(name))[0]
            o_type = git_ddls.get(name, sf_ddls.get(name))[1]
            o_file = git_ddls.get(name, (None, None, '! MISSING !'))[2][-30:]
            o_date = sf_ddls.get(name, (None, None, None))[2]
            o_date = f"{o_date:%Y-%m-%d %H:%M}" if o_date else '! MISSING !'

            logger.info(f"| {o_name:<39.39} | {o_type:<11.11} | {o_file:^30.30} | {o_date:^16.16} |")
    
    def compare_single_file(self, filename, branch=None):
        if branch is None:
            branch = self.sf_safe_branch
        if not self.EXTENSIONS.search(filename):
            logger.info(f"{filename} is not a SQL file. Running global comparison.")
            self.compare_sf_git(branch=branch)
            return
        
        o_name, o_type, easy_ddl, git_ddl = sql_meta(filename)
        sf_ddl = sf.get_ddl(branch, o_type, o_name)

        if not os.path.exists(self.DIFF_DIR):
            os.makedirs(self.DIFF_DIR)
        sf_filename = os.path.join(self.DIFF_DIR, os.path.basename(filename))
        with open(sf_filename, 'w') as sf_file:
            sf_file.write("-- Snowflake definition --\n")
            sf_file.write(statement_cleanup(sf_ddl))

        print_sql(git_ddl)
        print_sql(sf_ddl)


    def get_all_ddls(self):
        ddls = {}
        for f in Path(self.MODEL_DIR).rglob('*'):
            if self.EXTENSIONS.search(str(f)):
                o_name, o_type, easy_ddl, sql = sql_meta(str(f))
                ddls[(o_type + '#' + o_name).upper()] = (o_name, o_type, str(f))
        return ddls

    @staticmethod
    def _change_into_release_entry(change, commit_hash):
        """Converts a git.change of a file into a release candidate entry (string)."""
        change_type = change.change_type
        filename = change.a_path if change_type == 'D' else change.b_path
        if change_type != 'D':
            o_name, o_type, easy_ddl, sql = sql_meta(filename)

        if change_type == 'D':
            # dropped file/object
            return (f"--.File was removed and *will not* be included in the release.\n"
                    f"-- [{change_type}] NOT_INCLUDED:{filename}\n\n"
                    f"   You have to include a DROP statement to properly sync database state:\n"
                    f"   {release.HERE_STMT}\n\n")
        elif change_type == 'R':
            if easy_ddl:
                return (f"--.File with {o_type} definition was renamed and will be included in the release.\n"
                        f"-- [{change_type}] INCLUDED:{filename}\n\n"
                        f"   If you renamed the object, remember to drop the old one:\n"
                        f"   DROP {o_type} <<OLD_NAME>>; {release.HERE_STMT}\n\n")
            # rename table or stream
            return (f"--.File with {o_type} definition was renamed and *will not* be included in the release.\n"
                    f"-- [{change_type}] NOT_INCLUDED:{filename}\n\n"
                    f"   You may need to include an ALTER {o_type} RENAME statement to properly sync database state:\n"
                    f"   ALTER {o_type} <<OLD_NAME>> RENAME {o_name}; {release.HERE_STMT}\n\n")
        elif change_type == 'A':
            # add new object
            return Model._new_file_into_release_entry(change_type, easy_ddl, filename, o_name, o_type, sql)
        elif change_type == 'M':
            if easy_ddl:
                # modification / non table/stream
                return (f"--.File was changed and will be included in the release.\n"
                        f"--.You can change the order of the INCLUDE: statement below.\n"
                        f"-- [{change_type}] INCLUDED:{filename}\n\n")
            # modification / table/stream
            sql_diff = repo.get_file_diff(commit_hash, filename)
            return (f"--.File was changed but *will not be* included in the release.\n"
                    f"-- [{change_type}] NOT_INCLUDED:{filename}\n"
                    f"{sql_diff}\n"
                    f"   You may need to include an ALTER {o_type} statement to properly sync database state:\n"
                    +(f"   SELECT 1/IFF(COUNT(*)>0, 0, 1) FROM {o_name}; -- this will fail if stream is non empty\n" if o_type == 'STREAM' else '')
                    + f"   ALTER {o_type} {o_name} {release.HERE_STMT};\n\n")
        else:
            return (f"--.I don't understand this change and it it *will not* be included in the release.\n"
                    f"-- [{change_type}] NOT_INCLUDED:{filename}\n\n"
                    f"   You have to manually apply changes made in file listed above.\n"
                    f"   {release.HERE_STMT}\n\n")
    @staticmethod
    def _new_file_into_release_entry(change_type, easy_ddl, filename, o_name, o_type, sql):
        if easy_ddl:
            return (f"--.File was added and will be included in the release.\n"
                    f"--.You can change the order of the INCLUDE: statements below.\n"
                    f"-- [{change_type}] INCLUDED:{filename}\n\n")

        try:
            ddl = sf.get_ddl(repo.get_branch(), o_type, o_name)
        except RuntimeError as e:
            if 'does not exist' not in str(e):
                raise
            return (f"--.File with {o_type} definition was added and {o_name} does not exist on the server.\n"
                    f"--.It will be included in the release.\n"
                    f"--.You can change the order of the INCLUDE: statements below.\n"
                    f"-- [{change_type}] INCLUDED:{filename}\n\n")

        logger.warning(f"New CREATE {o_type} definition found in {filename}, but {o_name} already exists on the server.")
        sql_diff = get_diff_sql(ddl, sql, f"{o_type} {o_name} definition on the server",
                f"{o_type} {o_name} from {filename}")
        
        return (f"--.File with {o_type} definition was added but {o_name} *exists* on the server.\n"
                f"-- [{change_type}] NOT_INCLUDED:{filename}\n"
                f"{sql_diff}\n"
                f"   Double check if {o_type} definition on the server matches file contents. {release.HERE_STMT}\n"
                f"   SELECT '{o_type} {o_name} already exists on the server';\n\n")

model = Model()



import os

import hashlib

def get_file_contents(filename):
    """Returns file content as string."""
    with open(filename, 'r') as srcfile:
        return srcfile.read()

def hexDigest(string):
    """Returns md5 hash from s parameter."""
    return hashlib.md5(string.encode()).hexdigest()

def remove_file(filename):
    """Removes the file (if the file exist)."""
    if os.path.exists(filename):
        os.remove(filename)

def yes_or_no(question):
    while "the answer is invalid":
        reply = str(input(question+' (y/n): ')).lower().strip()
        if reply[:1] == 'y':
            return True
        if reply[:1] == 'n':
            return False



import re
from io import StringIO

import sqlparse
from termcolor import colored, cprint

from .utils import get_file_contents
from .log import logger

TABLE_DIR    = re.compile(r'/tables?/', re.I)
CREATE_TABLE = re.compile(r'create\s+((or\s+replace\s+)|(if\s+not\s+exists\s+))?table', re.I)

OR_REPLACE   = re.compile(r'create\s+or\s+replace\s+', re.I)
IF_NOT_EXISTS= re.compile(r'create\s+if\s+not\s+exists\s+', re.I)

OBJECT_TYPE  = (r'procedure|function|table|external\s+table|sequence|'
                r'view|materialized\s+view|file\s+format|stage|pipe|stream|task')
TYPE         = re.compile(OBJECT_TYPE, re.I)
TYPE_NAME    = re.compile(r'create\s+(or\s+replace\s+)?(if\s+not\s+exists\s+)?'
                          r'(local\s+|global\s+)?(temp\s+|temporary\s+|volatile\s+)?(transient\s+)?'
                          r'(?P<o_type>' + OBJECT_TYPE + r')\s+(?P<o_name>[\.\w-]+)', re.I)
DROP         = re.compile(r'drop\s+(' + OBJECT_TYPE + ')', re.I)

TYPE_DIR     = re.compile(r'/(?P<dir_type>' + OBJECT_TYPE.replace(r'\s+', '.') + r')s?/')

RESUME_TASK  = re.compile(r"alter\s+task\s+[\.\w-]+\s+resume\s*;",re.I)

_print_comment   = lambda x: cprint(x, 'green', attrs=['dark'], end='')
_print_keyword   = lambda x: cprint(x.upper(), 'blue', end='')
_print_other     = lambda x: cprint(x, 'white', end='')


def print_sql(sql):
    """Pretty prints SQL."""
    res = sqlparse.parse(sql)
    for r in res:
        for st in r:
            if type(st) == sqlparse.sql.Comment or str(st.ttype).startswith('Token.Comment'):
                _print_comment(st.value)
            elif st.is_keyword:
                _print_keyword(st.value)
            else:
                _print_other(st.value)
    print("\n")

def split_sql(sql):
    """Splits given SQL into list of statements, stripping comments."""
    sql = sqlparse.format(sql, strip_comments=True)
    return sqlparse.split(sql)

def sql_meta(filename):
    dir_type = TYPE_DIR.search(filename)
    assert dir_type, (f"Can't find valid object type prefix\nin {filename}")
    dir_type = dir_type.group('dir_type')

    sql = get_file_contents(filename)
    type_name = TYPE_NAME.search(sql)
    assert type_name, (f"Can't find a valid SQL CREATE statement\nin {filename}")

    o_type = type_name.group('o_type').upper().replace(' ', '_')
    o_name = type_name.group('o_name').upper()

    if dir_type.upper() != o_type:
        logger.warning(f"SQL CREATE {o_type} statement in folder named {dir_type}\nin{filename}")
    
    easy_ddl = False if o_type in ['TABLE', 'STREAM'] else True

    if not easy_ddl and not IF_NOT_EXISTS.search(sql):
        logger.debug(f"SQL CREATE {o_type} statement without 'IF NOT EXISTS' statement\nin {filename}")
    
    assert easy_ddl or not OR_REPLACE.search(sql), (
            f"Dangerous SQL CREATE {o_type} statement with 'OR REPLACE' statement\nin {filename}")
    
    assert easy_ddl or not DROP.search(sql), (
            f"Dangerous SQL DROP {o_type} statement\nin {filename}")

    if easy_ddl and not OR_REPLACE.search(sql):
        logger.warning(f"SQL CREATE {o_type} statement without 'OR REPLACE' statement\nin {filename}")
    
    return (o_name, o_type, easy_ddl, sql)

def get_diff_sql(left, right, fromfile, tofile):
    left = sqlparse.format(left, strip_comments=True)
    right = sqlparse.format(right, strip_comments=True)
    left = StringIO(left).readlines()
    right = StringIO(right).readlines()

    from difflib import unified_diff

    return "--.DIFF: " + "--.DIFF: ".join(unified_diff(left, right, fromfile=fromfile, tofile=tofile))        

def statement_cleanup(statement) -> str:
    """ Cleans up SQL statement. """
    statement = sqlparse.format(statement, strip_comments=True, keyword_case='lower',
                                identifier_case='lower')
    return statement 



import re

import snowflake.connector as sfc
from snowflake.connector.errors import Error as SfError
from snowflake.connector.errors import DatabaseError
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import dsa
from cryptography.hazmat.primitives import serialization

from .log import logger, is_debug
from .config import config
from .sql import split_sql, print_sql, RESUME_TASK
from .utils import yes_or_no

class Snowflake():
    """Snowflake connector wrapper."""

    SF_VALID_NAME = re.compile(r'^[\w-]+$')
    SF_PROD_NAME = config.read_config('production_db')
    SF_STAGING_NAME = config.read_config('staging_db')

    def __init__(self):
        self._conn = None

    def connect(self, branch):
        """Connect to the DB and returns connection."""
        db = self.get_db_name(branch)
        key = self._get_key(config.read_user_config('private_key_file'))
        user = config.read_user_config('user')
        warehouse = config.read_config('warehouse')
        role = config.read_config('role')
        logger.debug("Connecting to Snowflake database {} as {}".format(db,
            user))
        logger.debug("Role {}, warehouse {}".format(role, warehouse))
        try:
            if not key:
                logger.warning('Connecting to Snowflake using password. Please use key-pair auth instead.')
                return sfc.connect(password=config.read_user_config('password'),
                        user=user,
                        account=config.read_config('account'),
                        warehouse=warehouse,
                        database=db,
                        autocommit=False,
                        role=role,
                        validate_default_parameters=True,
                        schema='PUBLIC')
            return sfc.connect(private_key=key,
                    user=user,
                    account=config.read_config('account'),
                    warehouse=warehouse,
                    database=db,
                    autocommit=False,
                    role=role,
                    #validate_default_parameters=True,
                    schema='PUBLIC')
        except DatabaseError as de:
            if "250001 (08001)" in str(de):
                logger.info("Is this your first run in this branch and the database was not cloned? Try 'clone' first.")
            raise RuntimeError(de)

    def perform_release(self, sql, branch):
        """Run arbitrary SQL statement(s)."""
        conn = self.connect(branch)
        skip_resume_task = (self.get_db_name(branch) != self.SF_PROD_NAME)
        cur = conn.cursor()
        try:
            cur.execute(config.sql('autocommit'))
            cur.execute(config.sql('transaction_abort'))
            logger.debug('BEGIN TRANSACTION')
            cur.execute(config.sql('transaction_begin'))
            for statement in split_sql(sql):
                if skip_resume_task and RESUME_TASK.search(statement):
                    logger.info("Skipping '{}' statement as this is not production".format(
                        statement.replace("\n", " ")))
                    continue
                logger.debug('  running statement:')
                if is_debug():
                    print_sql(statement)
                try:
                    cur.execute(statement)
                except SfError as e:
                    if 'Empty SQL statement' in str(e):
                        logger.warning("Found empty SQL statement. Too many ';' in file?")
                    else:
                        logger.error("Release failed due to this statement:")
                        print_sql(statement)
                        raise
                if cur.rowcount:
                    logger.info("  " + str(cur.fetchone())[1:-1])
            logger.debug('COMMIT TRANSACTION')
            cur.execute(config.sql('commit'))
        except SfError as e:
            logger.debug('ROLLBACK TRANSACTION')
            conn.rollback()
            logger.error('Error while running the release:')
            raise RuntimeError(e)
        finally:
            conn.close()

    def run_single_statament(self, query, branch='main'):
        """Runs single SQL statement against Snowflake."""
        conn = self.connect(branch)
        cur = conn.cursor()
        try:
            if is_debug():
                print_sql(query)
            cur.execute(query)
            conn.commit()
            return cur.fetchall()
        except SfError as e:
            conn.rollback()
            if "This session does not have a current database" in str(e):
                logger.info("Is this your first run in this branch and the database was not cloned? Try 'clone' first.")
            raise RuntimeError(e)
        finally:
            conn.close()

    def clone_production(self, branch, force=False):
        """Clones production database into new db named after branch name."""
        newdb = self.get_db_name(branch)
        assert self.SF_VALID_NAME.match(newdb), (f"{newdb} is not a valid Snowflake"
                " identifier")
        assert newdb.strip().upper() != self.SF_PROD_NAME, ("Trying to create new database"
                f" with name {self.SF_PROD_NAME} (which is production db name!)")     
        assert newdb.strip().upper() != self.SF_STAGING_NAME or force, (
                f"Trying to recreate {self.SF_STAGING_NAME} without --force.")
        
        if not force:
            logger.info(f"Checking if {newdb} already exists...")
            sql = config.sql('clone_exists').format(newdb=newdb)
            clone_exists = self.run_single_statament(sql)
        
            if clone_exists[0][0] != 0 and not yes_or_no(f"{newdb} already exists. Are you sure you want to replace it"):
                return

        logger.info(f"Cloning {self.SF_PROD_NAME} into {newdb}")

        sql = config.sql('create_clone').format(newdb=newdb, prod=self.SF_PROD_NAME)
        self.run_single_statament(sql)
        logger.info("Cloning finished")
    
    def drop_clone(self, branch, force=False):
        """Drops clone."""
        db = self.get_db_name(branch)
        assert self.SF_VALID_NAME.match(db), (f"{db} is not a valid Snowflake"
                " identifier")
        assert db.strip().upper() != self.SF_PROD_NAME, ("Never ever drop production!")     
        assert db.strip().upper() != self.SF_STAGING_NAME or force, (
                f"Trying to drop {self.SF_STAGING_NAME} without --force.")
        
        logger.info(f"Dropping clone {db}")

        self.run_single_statament(f'DROP DATABASE {db};')
        logger.info("Dropping clone finished")

    def get_dev_clones(self):
        """Returns all development clones""" 
        sql = config.sql('get_dev_clones')
        return self.run_single_statament(sql)
    
    def get_altered_objects(self, branch):
        """Returns all objects changed since clone creation."""
        db = self.get_db_name(branch)
        sql = config.sql('get_altered_objects').format(db=db)
        return self.run_single_statament(sql, branch)
    
    def get_all_objects(self, branch):
        db = self.get_db_name(branch)
        sql = config.sql('get_all_objects').format(db=db)
        all_objects = self.run_single_statament(sql, branch)

        sql = config.sql('get_streams').format(db=db)
        streams = self.run_single_statament(sql, branch)
        for stream in streams:
            all_objects += [(stream[3], stream[1], 'STREAM', stream[0])]

        sql = config.sql('get_tasks').format(db=db)
        tasks = self.run_single_statament(sql, branch)
        for task in tasks:
            all_objects += [(task[4], task[1], 'TASK', task[0])]
        
        dictionary = {}
        for obj in all_objects:
            dictionary[(obj[2].replace(' ', '_') + '#' + obj[0] + '.' + obj[1]).upper()] = (obj[0] + '.' + obj[1], obj[2], obj[3])

        return dictionary
    
    def get_ddl(self, branch, o_type, o_name) -> str:
        """Returns object DDL."""
        if o_type.lower() == 'stage':
            return f'Unable to get DDL for stage {o_name}'
        sql = config.sql('get_ddl').format(o_type=o_type, o_name=o_name,
                                           parameters="()" if o_type=="PROCEDURE" else "")
        return self.run_single_statament(sql, branch)[0][0]

    def get_db_name(self, branch):
        """Converts branch name into database name to operate on."""
        if(branch.upper() == 'MAIN'):
            return self.SF_PROD_NAME
        elif(branch.upper() == 'DEVELOP'):
            return self.SF_STAGING_NAME
        else:
            return '_DEV_' + branch.upper()[:30]

    def _get_key(self, keypath):
        """Converts PKCS8 file into bytes array acceptable by Snowflake connector."""
        with open(keypath, "rb") as key:
            p_key = serialization.load_pem_private_key(
                    key.read(),
                    password=None,
                    backend=default_backend())
            return p_key.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption())

sf = Snowflake()



import os
import re
from datetime import datetime

from .config import config
from .log import logger, is_debug
from .utils import get_file_contents, hexDigest, remove_file
from .dwhrepo import repo
from .snowflake import sf
from .sql import sql_meta, print_sql

class Release():
    """Performs releases and handles release files."""
    RELEASES_DIR       = config.read_config('releases_dir', default='releases')
    MODEL_DIR          = config.read_config('model_dir', default='model')
    RELEASE_CANDIDATE  = os.path.join(RELEASES_DIR, 'release_candidate.sql')
    RELEASE_SHA        = os.path.join(RELEASES_DIR, 'release_candidate.sha')
    RELEASE_TABLE      = 'PUBLIC.DWH_RELEASES_HISTORY'
    INCLUDED           = re.compile(r'-- \[(?P<change>.)\] (?P<inc>(NOT_)?INCLUDED):(?P<file>{}/\S+)( #)?(?P<hash>\S+)?'.format(MODEL_DIR))
    HERE_STMT          = '<<HERE>>'

    def __init__(self):
        self.sf_safe_branch  = repo.get_sf_safe_branch()

    def check_release_candidate(self):
        """Asserts if release candidate file existis and was not modified."""
        if self.release_candidate_exists():
            if self.release_candidate_modified():
                raise RuntimeError("{} was changed or you changed branch, can't create new "
                        "release candidate.\nEither remove it, allow me to do so by adding --force flag, or deploy it.".format(self.RELEASE_CANDIDATE))
            logger.info("Release candidate file exists, but it was not modified. Will be replaced.")

    def release_candidate_modified(self):
        """Returns bool if release candidate file was manually modified."""
        if not os.path.exists(self.RELEASE_SHA):
            raise RuntimeError("Release candidate file {} exists but MD5 sum file is missing.".format(self.RELEASE_CANDIDATE))
        branch = repo.get_branch()
        rc_file_hash = hexDigest(get_file_contents(self.RELEASE_CANDIDATE) + branch)
        rc_stored_md5 = get_file_contents(self.RELEASE_SHA)
        
        return rc_file_hash != rc_stored_md5

    def get_release_candidate_lines(self):
        """Returns release candidate file contents as lines."""
        with open(self.RELEASE_CANDIDATE, 'r') as release_candidate_file:
            return release_candidate_file.readlines()
    
    def check_release_dir_clean(self, base_commit) -> None:
        """Checks if all the release files were applied in current DB."""
        files_added = self.get_unsynced_releases(base_commit)

        if len(files_added) > 0:
            raise RuntimeError("Files present in {} folder that were not applied on the "
                    "database.\nYou have to 'sync' first.".format(self.RELEASES_DIR))

    def sync(self, branch=None, dry_run=False) -> None:
        """Syncs non-applied changes in releases and model folders."""
        if branch is None:
            branch = self.sf_safe_branch

        commit_hash = self.get_base_commit(branch)
        files = self.get_unsynced_releases(commit_hash)

        if files.values():
            logger.warning("Syncing changes:")
        else:
            logger.info("No pending changes to sync.")
        for change in files.values():
            if change.change_type == 'D':
                logger.warning(f"Skipping removed release file {change.a_path}.")
                continue

            changed_file = change.b_path
            logger.info("Running release file {}:".format(changed_file))

            deploy_sql = self.release_file_to_sql(get_file_contents(changed_file))

            if is_debug():
                print_sql(deploy_sql)
            if dry_run:
                if not is_debug():
                    print_sql(deploy_sql)
                logger.info("Skipping SQL execution due to --dry-run.")
                return

            sf.perform_release(deploy_sql, branch)
            self.insert_release_entry(changed_file, branch)
    
    def test_sync(self):
        last_commit_sha = repo.get_last_commit_sha()
        sf.clone_production(branch=last_commit_sha, force=True)
        try:
            release.sync(branch=last_commit_sha)
        except Exception as e:
            raise
        finally:
            sf.drop_clone(branch=last_commit_sha)

    def save_release(self, sql):
        """Save release file (if needed) commits it and adds new entry in
           DWH changelog table."""

        release_filename = self._gen_release_filename()
        logger.info(f'Writing to {release_filename}.')
        with open(release_filename, 'a+') as release_file:
            release_file.write(sql)

        repo.commit_release(release_filename)

        self.insert_release_entry(release_filename, self.sf_safe_branch)
        self.remove_release_candidate()

    def remove_release_candidate(self):
        """Removes release candidate files."""
        if self.release_candidate_modified():
            logger.warning(f'Removing modified release candidate file {self.RELEASE_CANDIDATE}.')
        else:
            logger.info(f'Removing release candidate file {self.RELEASE_CANDIDATE}.')

        remove_file(self.RELEASE_CANDIDATE)
        remove_file(self.RELEASE_SHA)

    def release_candidate_exists(self):
        """Checks if release candidate file exists."""
        return os.path.exists(self.RELEASE_CANDIDATE)

    def save_release_candidate_file(self, sql, branch):
        """Saves release candidate file and MD5 sum file."""
        with open(self.RELEASE_CANDIDATE, 'w') as release_candidate_file:
            release_candidate_file.write(sql)
        with open(self.RELEASE_SHA, 'w') as release_candidate_md5:
            # by including branch in hash we ensure noone will run release candidate
            # from branch "A" after switching to branch "B"
            release_candidate_md5.write(hexDigest(sql + branch))
        
        logger.info(f"{self.RELEASE_CANDIDATE} and {self.RELEASE_SHA} created.")

    def print_release_history(self):
        """Returns release history."""
        release_history = self.release_history(self.sf_safe_branch)
        logger.info("|{:_^13}|{:_^32.32}|{:_^22}|{:_^18}|{:_^6}|{:_^6}|".format(
            'commit hash', 'file name', 'applied by', 'applied on', 'branch', 'prod'))
        for rel in release_history:
            logger.info("| {:>10.10}  | {:<30.30} | {:^20.20} | {:%Y-%m-%d %H:%M} | {:^4} | {:^4} |".format(
                rel[0], str(rel[1])[-30:], rel[2], rel[3], '•' if rel[4] else '', '•' if rel[5] else ''))

    def release_history(self, branch):
        """Returns release history from DB."""
        sql = config.sql('release_history').format(
            RELEASE_TABLE=self.RELEASE_TABLE, SF_PROD_NAME=sf.SF_PROD_NAME)
        history = sf.run_single_statament(sql, branch)
        return history

    def get_base_commit(self, branch):
        """Queries the based commit in release history table."""
        
        sql = config.sql('get_base_commit').format(RELEASE_TABLE=self.RELEASE_TABLE)
        commit = sf.run_single_statament(sql, branch)
        assert len(commit) > 0, "No data in release history table, you need to create an initial entry"
        logger.debug("Base commit hash in current database is {}."
                .format(commit[0][0]))
        return commit[0][0]

    def insert_release_entry(self, filename, branch):
        """Inserts a row in a release history table."""
        commit = repo.get_file_last_commit_hash(filename)

        sql = config.sql('insert_release_entry').format(
            RELEASE_TABLE=self.RELEASE_TABLE, commit=commit, filename=filename)
        sf.run_single_statament(sql, branch)
        logger.info("New entry in release history table with {}"
                .format(commit))

    def release_candidate_contains(self, contains):
        """Checks if specific entry is present in the release history file."""
        file_contents = get_file_contents(self.RELEASE_CANDIDATE)
        return contains in file_contents

    def prepare_release_file(self):
        """Deploys a release based on existing release candidate file."""
        if not self.release_candidate_exists():
            raise RuntimeError("Release candidate file does not exist, can't "
                    "make a release. Prepare release first.")

        if self.release_candidate_contains(self.HERE_STMT):
            raise RuntimeError(f"Code placeholder ({self.HERE_STMT}) found in the release "
                    "candidate file. Replace it with a valid SQL statement "
                    "or remove the line.")

        branch = repo.get_branch()
        user = config.read_user_config('user')

        release_sql = "\n-- RELEASE FROM BRANCH {}\n-- {} on {:%Y-%m-%d %H:%M}\n\n".format(
            branch, user, datetime.now())

        for line in release.get_release_candidate_lines():
            if line.startswith('--.'):
                continue
        
            included = self.INCLUDED.search(line)    
            if included and included.group('change') != 'D':
                release_sql += line.strip()
                release_sql += repo.get_file_last_commit(included.group('file'))
                continue
            
            release_sql += line
        
        return release_sql

    def release_file_to_sql(self, release_sql) -> str:
        """Converts release file (with filenames) to SQL."""
        sql = ""
        for line in release_sql.splitlines():
            included = self.INCLUDED.search(line)
            logger.debug(line)
            if included:
                sql += line + "\n"
                if included.group('inc') == 'INCLUDED':
                    sql += repo.get_file_contents_by_commit(included.group('file'), included.group('hash'))

            if not line.startswith('--'):
                sql += line + "\n"

        return sql
    
    def get_unsynced_releases(self, base_commit):
        """Returns list with unsynced releases."""
        return repo.get_changed_files(index=base_commit,
                prefix=self.RELEASES_DIR, change_type=('A', 'R', 'M'))
    
    def compare_branches_and_clones(self) -> None:
        branches = repo.get_dev_branches()
        branches = list(map(lambda x: sf.get_db_name(x), branches))
        
        clones = sf.get_dev_clones()
        logger.info("|{:_^30}|{:_^18}|{:_^18}|{:_^14}|".format(
            'clone', 'created', 'updated', 'has branch?'))
        for clone in clones:
            logger.info("| {:<28.28} | {:%Y-%m-%d %H:%M} | {:%Y-%m-%d %H:%M} | {:^12} |".format(
                clone[0], clone[1], clone[2], 'yes' if clone[0] in branches else 'no'))

    def _gen_release_filename(self) -> str:
        """Generates release filename based on branch."""
        base = repo.get_branch()
        filebase = os.path.join(self.RELEASES_DIR, base)
        return f"{filebase}.sql"


release = Release()



import configparser
import functools

from os import path
from pathlib import Path

class Config():
    """Config wrapper that reads global config and user config."""

    PROJECT_ROOT = path.join(path.dirname(path.realpath(__file__)), '..')
    CONFIG_INI = path.join(PROJECT_ROOT, 'config.ini')
    HOME_DIR = Path.home()
    CONN_INI = path.join(HOME_DIR, '.snowflake-cicd.ini')

    def __init__(self):
        pass
    
    def __lazy_init(func):
        """Reads and parses global config file and user's conn file."""
        @functools.wraps(func)
        def wrap(self, *args, **kwargs):
            if not hasattr(self, '_config'):
                assert path.exists(Config.CONFIG_INI), f"Missing config file at path {Config.CONFIG_INI}"
                self._config = configparser.ConfigParser()
                self._config.read(Config.CONFIG_INI)

            if not hasattr(self, '_conn'):
                assert path.exists(Config.CONN_INI), f"Missing connection settings file at path {Config.CONN_INI}"
                self._conn = configparser.ConfigParser()
                self._conn.read(Config.CONN_INI)

            return func(self, *args, **kwargs)
        return wrap

    @__lazy_init
    def read_config(self, key, section='default', default=None) -> str:
        """Reads [section] key from user's conn file or use global file
           if the key is missing."""
        return self._conn[section].get(key,
               self._config[section].get(key, default))

    @__lazy_init
    def read_user_config(self, key, section='default', default=None) -> str:
        """Reads [section] from user .snowflake-cicd.ini file."""
        return self._conn[section].get(key, default)
    
    @__lazy_init
    def sql(self, query_id) -> str:
        """Returns value from config section 'queries'."""
        return self._config['queries'].get(query_id)

config = Config()



import os
import re
from datetime import datetime

from git import Repo, InvalidGitRepositoryError

from .log import logger
from .config import config


class DWHRepo(Repo):
    """Wrapper class git git.Repo to handle DWH repository specific tasks."""
    MODEL_DIR = None
    SF_SAFE   = re.compile(r'\W')

    def __init__(self):
        """Inits the repo from parent folder and initialise _tags"""
        super().__init__(search_parent_directories=True)
        DWHRepo.MODEL_DIR = config.read_config('model_dir', default='model')
        self._tags = sorted(self.tags, key=lambda t: t.commit.committed_datetime)

    def get_branch(self, filename_safe=True):
        """Returns active branch name as string."""
        if filename_safe:
            return self.active_branch.name.replace('/', '_')
        return self.active_branch.name

    def get_sf_safe_branch(self, branch=None):
        """Returns branch name in format safe for Snowflake object names."""
        if not branch:
            branch = self.active_branch.name
        return self.SF_SAFE.sub('_', branch).upper()

    def get_changed_files(self, index='main', prefix='', suffix='.sql',
            change_type="*"):
        """Returns all files changed from index."""
        files = {}
        for d in self.commit(index).diff():
            if d.b_path.startswith(prefix) and d.b_path.endswith(suffix) and (d.change_type in change_type or change_type == '*'):
                files[d.b_path] = d
                logger.info("  [{}] {}".format(d.change_type, d.b_path))
        return files
    
    def get_file_contents_by_commit(self, filename, commit):
        return self.git.show(f"{commit}:{filename}")

    def get_file_last_commit(self, infile):
        """Returns a three lines string description of the last change made to the file."""
        return self.git.log('-1', '--pretty= #%h%n-- change on:   %ai by %an: %s%n'
                f'-- show file:   git show %h:{infile}%n'
                f'-- last change: git diff %h^! {infile}%n', infile)
    
    def get_file_diff(self, commit, infile):
        diff = self.git.diff(commit, '--unified=0', '--abbrev', '--ignore-space-change', '--no-prefix', infile)

        diff = "\n--.DIFF: ".join(_t for _t in diff.split("\n") if not infile in _t and not _t.startswith('index ') )
        return "--.DIFF: " + diff + "\n"
    
    def diff_from_prod(self):
        diff = self.git.diff('main', '--color=always', self.MODEL_DIR)
        diff = list(filter(lambda x: not '--- a/' in x, diff.split("\n")))
        diff = list(filter(lambda x: not '+++ b/' in x, diff))
        print("\n".join(diff))

    def get_file_last_commit_hash(self, infile):
        """Returns last commit hash of the last change made to the file."""
        return self.git.log('-1', '--pretty=%H', infile)

    def assert_repo(self):
        """Checks if the repo is dirty, and if this can be allowed."""
        branch = self.get_branch()

        if self.is_dirty():
            if not self.is_model_clean():
                raise RuntimeError(f'{self.MODEL_DIR} not clean! Commit your changes.')

    def commit_release(self, release_file):
        """Adds release_file to stage and commits it."""
        self.index.add([release_file])
        self.index.commit('(DWH new release)')
        try:
            self.remote(name='origin').push()
        except (ValueError):
            logger.error('Can\'t push changes to remote!')

        
    def is_model_clean(self):
        """Logs all the changed (uncommited) files in repo."""
        model_clean = True
        if self.untracked_files:
            logger.warning('Untracked files:')
            for item in self.untracked_files:
                logger.warning('  - ' + item)
                model_clean = model_clean and not item.startswith(self.MODEL_DIR)
        if self.index.diff(None):
            logger.warning('Not commited files:')
            for item in self.index.diff(None):
                logger.warning('  - ' + item.b_path)
                model_clean = model_clean and not item.b_path.startswith(self.MODEL_DIR)
        return model_clean

    def get_dev_branches(self):
        """Returns a list of development branches and."""
        self.git.fetch("--all")
        branches = []
        for branch in self.branches:
            branches.append(self.get_sf_safe_branch(branch.name))
        return branches
    
    def get_last_commit_sha(self) -> str:
        """Returns last commit SHA (40 chars)"""
        return self.head.commit.hexsha

try:
    repo = DWHRepo()
except InvalidGitRepositoryError:
    logger.error('Run cicd tool inside a data model GIT repository. See README for details.')
    quit(-1)
except AssertionError as e:
    logger.error(e)
    quit(-1)
